@@ -28,11 +28,14 @@
 #include <sys/proc.h>
 #include <sys/stack.h>
 #include <sys/sysproto.h>
+#include <sys/queue.h>
+#include <sys/mutex.h>
 
 #include "nvkms-ioctl.h"
 #include "nvidia-modeset-os-interface.h"
 #include "nvkms.h"
 #include "nv-modeset-interface.h"
+#include "nv-kthread-q.h"
 
 #include "nv-retpoline.h"
 
@@ -109,22 +112,62 @@ struct nvkms_per_open {
             struct selinfo select;
         } user;
         struct {
-            nvkms_timer_handle_t *task;
-            struct NvKmsKapiDevice *device;
+            nv_kthread_q_item_t nv_kthread_q_item;
         } kernel;
     };
 };
 
+NvBool nvkms_test_fail_alloc_core_channel(enum FailAllocCoreChannelMethod method)
+{
+    return NV_FALSE;
+}
+
 NvBool nvkms_output_rounding_fix(void)
 {
     return NV_TRUE;
 }
 
+NvBool nvkms_disable_hdmi_frl(void)
+{
+    return NV_FALSE;
+}
+
 NvBool nvkms_disable_vrr_memclk_switch(void)
 {
     return NV_FALSE;
 }
 
+NvBool nvkms_hdmi_deepcolor(void)
+{
+    return NV_TRUE;
+}
+
+NvBool nvkms_vblank_sem_control(void)
+{
+    return NV_TRUE;
+}
+
+NvBool nvkms_opportunistic_display_sync(void)
+{
+    return NV_TRUE;
+}
+
+enum NvKmsDebugForceColorSpace nvkms_debug_force_color_space(void)
+{
+    return NVKMS_DEBUG_FORCE_COLOR_SPACE_NONE;
+}
+
+NvBool nvkms_enable_overlay_layers(void)
+{
+    return NV_TRUE;
+}
+
+/*************************************************************************
+ * Used for queueing nvkms events. We do not use the nvkms timer abstraction
+ * since it locks nvkms_lock
+ *************************************************************************/
+static struct nv_kthread_q nvkms_kthread_q;
+
 /*************************************************************************
  * nvidia-modeset-os-interface.h functions.  It is assumed that these
  * are called while nvkms_lock is held.
@@ -229,6 +272,11 @@ NvBool nvkms_syncpt_op(
     return NV_FALSE;
 }
 
+NvBool nvkms_kernel_supports_syncpts(void)
+{
+    return NV_FALSE;
+}
+
 int nvkms_snprintf(char *str, size_t size, const char *format, ...)
 {
     int ret;
@@ -542,13 +590,9 @@ static void nvkms_resume(NvU32 gpuId)
     sx_xunlock(&nvkms_lock);
 }
 
-static void nvkms_kapi_task_callback(void *arg, NvU32 dataU32 __unused)
+static void nvkms_kapi_task_callback(void *arg)
 {
-    struct nvkms_per_open *popen = arg;
-    struct NvKmsKapiDevice *device = popen->kernel.device;
-
-    nvkms_free_timer(popen->kernel.task);
-    popen->kernel.task = NULL;
+    struct NvKmsKapiDevice *device = arg;
 
     nvKmsKapiHandleEventQueueChange(device);
 }
@@ -569,11 +613,10 @@ nvkms_event_queue_changed(nvkms_per_open_handle_t *pOpenKernel,
 
             break;
         case NVKMS_CLIENT_KERNEL_SPACE:
-            if (!popen->kernel.task) {
-                popen->kernel.task = nvkms_alloc_timer(nvkms_kapi_task_callback,
-                                                       popen,
-                                                       0, /* dataU32 */
-                                                       0 /* callout delay */);
+            if (eventsAvailable) {
+                nv_kthread_q_schedule_q_item(
+                    &nvkms_kthread_q,
+                    &popen->kernel.nv_kthread_q_item);
             }
             break;
     }
@@ -823,11 +866,9 @@ static struct nvkms_per_open *nvkms_open_common(enum NvKmsClientType type,
             break;
         case NVKMS_CLIENT_KERNEL_SPACE:
             /* enqueue our new task */
-            popen->kernel.device = device;
-            popen->kernel.task = nvkms_alloc_timer(nvkms_kapi_task_callback,
-                                                   popen,
-                                                   0, /* dataU32 */
-                                                   0 /* callout delay */);
+            nv_kthread_q_item_init(&popen->kernel.nv_kthread_q_item,
+                                   nvkms_kapi_task_callback,
+                                   device);
             break;
     }
 
@@ -855,9 +896,9 @@ static void nvkms_close_common(struct nvkms_per_open *popen)
 {
     sx_xlock(&nvkms_lock);
     nvKmsClose(popen->data);
+    popen->data = NULL;
     sx_xunlock(&nvkms_lock);
 
-    mtx_destroy(&popen->lock);
     switch (popen->type) {
         case NVKMS_CLIENT_USER_SPACE:
             break;
@@ -876,11 +917,11 @@ static void nvkms_close_common(struct nvkms_per_open *popen)
              * attempts to call back into NVKMS, the popen->data==NULL check in
              * nvkms_ioctl_common() should reject the request.
              */
-            nvkms_free_timer(popen->kernel.task);
-            popen->kernel.task = NULL;
+            nv_kthread_q_flush(&nvkms_kthread_q);
             break;
     }
 
+    mtx_destroy(&popen->lock);
     nvkms_free(popen, sizeof(*popen));
 
     sx_xlock(&nvkms_module.lock);
@@ -937,6 +978,15 @@ NvBool nvkms_ioctl_from_kapi
             (NvU64)(NvUPtr)params_address, params_size) == 0;
 }
 
+NvBool nvkms_ioctl_from_kapi_try_pmlock
+(
+    struct nvkms_per_open *popen,
+    NvU32 cmd, void *params_address, const size_t params_size
+)
+{
+    /* FreeBSD doesn't have a power management lock */
+    return nvkms_ioctl_from_kapi(popen, cmd, params_address, params_size);
+}
 
 /*************************************************************************
  * APIs for locking.
@@ -981,6 +1031,41 @@ NvBool nvKmsKapiGetFunctionsTable(struct NvKmsKapiFunctionsTable *funcsTable)
     return nvKmsKapiGetFunctionsTableInternal(funcsTable);
 }
 
+NvU32 nvKmsKapiF16ToF32(NvU16 a)
+{
+    return nvKmsKapiF16ToF32Internal(a);
+}
+
+NvU16 nvKmsKapiF32ToF16(NvU32 a)
+{
+    return nvKmsKapiF32ToF16Internal(a);
+}
+
+NvU32 nvKmsKapiF32Mul(NvU32 a, NvU32 b)
+{
+    return nvKmsKapiF32MulInternal(a, b);
+}
+
+NvU32 nvKmsKapiF32Div(NvU32 a, NvU32 b)
+{
+    return nvKmsKapiF32DivInternal(a, b);
+}
+
+NvU32 nvKmsKapiF32Add(NvU32 a, NvU32 b)
+{
+    return nvKmsKapiF32AddInternal(a, b);
+}
+
+NvU32 nvKmsKapiF32ToUI32RMinMag(NvU32 a, NvBool exact)
+{
+    return nvKmsKapiF32ToUI32RMinMagInternal(a, exact);
+}
+
+NvU32 nvKmsKapiUI32ToF32(NvU32 a)
+{
+    return nvKmsKapiUI32ToF32Internal(a);
+}
+
 /*************************************************************************
  * File operation callback functions.
  *************************************************************************/
@@ -1115,7 +1200,7 @@ static int nvkms_linux_ioctl_function(
     struct linux_ioctl_args *args
 )
 {
-    uint32_t dir;
+    uint32_t dir = 0;
 
     /*
      * Linux allocates 14 bits for the parameter size, whereas FreeBSD only
@@ -1209,6 +1294,13 @@ nvidia_modeset_loader(struct module *m, int what, void *arg)
             return ret;
         }
 
+        ret = nv_kthread_q_init(&nvkms_kthread_q,
+                "nvidia-modeset/kthread_q");
+        if (ret != 0) {
+            nvkms_free_rm();
+            return ENOMEM;
+        }
+
         /*
          * nvkms_lock will be recursed on, specifically through the nvkms
          * taskqueue callback. This will lock nvkms_lock, then dispatch the
@@ -1231,6 +1323,7 @@ nvidia_modeset_loader(struct module *m, int what, void *arg)
                              "nvidia-modeset");
 
         if (nvkms_dev == NULL) {
+            nv_kthread_q_stop(&nvkms_kthread_q);
             sx_destroy(&nvkms_module.lock);
             mtx_destroy(&nvkms_timers.lock);
             sx_destroy(&nvkms_lock);
@@ -1327,6 +1420,8 @@ nvidia_modeset_loader(struct module *m, int what, void *arg)
 
         taskqueue_run(taskqueue_nvkms);
 
+        nv_kthread_q_stop(&nvkms_kthread_q);
+
         destroy_dev(nvkms_dev);
         nvkms_dev = NULL;
 
