@@ -1,6 +1,6 @@
 /* _NVRM_COPYRIGHT_BEGIN_
  *
- * Copyright 2001-2022 by NVIDIA Corporation.  All rights reserved.  All
+ * Copyright 2001-2024 by NVIDIA Corporation.  All rights reserved.  All
  * information contained herein is proprietary and confidential to NVIDIA
  * Corporation.  Any use, reproduction, or disclosure without the written
  * permission of NVIDIA Corporation is prohibited.
@@ -8,10 +8,22 @@
  * _NVRM_COPYRIGHT_END_
  */
 
+// This must be defined before #including nv-firmware.h (which may be included
+// indirectly by some headers below).
+#define NV_FIRMWARE_FOR_NAME(name)  "nvidia_" name "_fw"
+
 #include "os-interface.h"
 #include "nv.h"
 #include "nv-freebsd.h"
 #include "nv-caps.h"
+#include "nv-firmware.h"
+
+#include <sys/consio.h>
+#include <sys/fbio.h>
+#include <sys/linker.h>
+#include <sys/firmware.h>
+#include <dev/syscons/syscons.h>
+#include <machine/metadata.h>
 
 uma_zone_t nvidia_stack_t_zone;
 static nvidia_stack_t *__nvidia_init_sp = NULL;
@@ -20,7 +32,7 @@ devclass_t nvidia_devclass;
 nv_state_t nvidia_ctl_state;
 nv_cap_t *nvidia_caps_root = NULL;
 
-const NvBool nv_is_rm_firmware_supported_os = NV_FALSE;
+const NvBool nv_is_rm_firmware_supported_os = NV_TRUE;
 
 static struct callout snapshot_timer;
 static struct mtx snapshot_timer_mutex;
@@ -61,38 +73,6 @@ int nvidia_attach(device_t dev)
     if ((rm_is_supported_device(sc->attach_sp, nv)) != NV_OK)
         return ENXIO;
 
-    for (i = 0; i < NV_GPU_NUM_BARS; i++) {
-        if (sc->BAR_recs[i] != NULL) {
-            sc->BAR_sg_lists[i] = sglist_alloc(1, M_WAITOK);
-            if (!sc->BAR_sg_lists[i])
-                goto failed;
-
-            sglist_append_phys(sc->BAR_sg_lists[i],
-                    nv->bars[i].cpu_address, nv->bars[i].size);
-
-            sc->BAR_objects[i] = vm_pager_allocate(OBJT_SG,
-                    sc->BAR_sg_lists[i],
-                    nv->bars[i].size, (VM_PROT_READ | VM_PROT_WRITE),
-                    0, NULL);
-            if (!sc->BAR_objects[i])
-                goto failed;
-
-            VM_OBJECT_WLOCK(sc->BAR_objects[i]);
-            switch (i) {
-                case NV_GPU_BAR_INDEX_FB:
-                    vm_object_set_memattr(sc->BAR_objects[i],
-                            VM_MEMATTR_WRITE_COMBINING);
-                    break;
-                case NV_GPU_BAR_INDEX_REGS:
-                default:
-                    vm_object_set_memattr(sc->BAR_objects[i],
-                            VM_MEMATTR_UNCACHEABLE);
-                    break;
-            }
-            VM_OBJECT_WUNLOCK(sc->BAR_objects[i]);
-        }
-    }
-
     sc->dma_mask = 0xffffffffULL;
 
     if ((status = nvidia_dev_attach(sc)) != 0)
@@ -106,25 +86,12 @@ int nvidia_attach(device_t dev)
     nv_sysctl_init(nv);
 
     return 0;
-
-failed:
-    for (i = 0; i < NV_GPU_NUM_BARS; i++) {
-        if (sc->BAR_recs[i] != NULL) {
-            if (sc->BAR_objects[i])
-                NV_VM_OBJECT_DEALLOCATE(sc->BAR_objects[i]);
-            if (sc->BAR_sg_lists[i])
-                NV_SGLIST_FREE(sc->BAR_sg_lists[i]);
-        }
-    }
-
-    return ENOMEM;
 }
 
 int nvidia_detach(device_t dev)
 {
     int status;
     struct nvidia_softc *sc;
-    uint32_t i;
 
     sc = device_get_softc(dev);
     nv_sysctl_exit(sc->nv_state);
@@ -141,15 +108,6 @@ int nvidia_detach(device_t dev)
         goto failed;
     }
 
-    for (i = 0; i < NV_GPU_NUM_BARS; i++) {
-        if (sc->BAR_recs[i] != NULL) {
-            if (sc->BAR_objects[i])
-                NV_VM_OBJECT_DEALLOCATE(sc->BAR_objects[i]);
-            if (sc->BAR_sg_lists[i])
-                NV_SGLIST_FREE(sc->BAR_sg_lists[i]);
-        }
-    }
-
 failed:
     /* XXX Fix me? (state) */
     return status;
@@ -369,6 +327,22 @@ int nvidia_handle_ioctl(
     int status;
     int nr, size;
 
+    /*
+     * Treat FIONBIO and FIOASYNC with data == 0 as a no-op. This allows fcntl
+     * F_SETFL to work (if O_NONBLOCK and O_ASYNC are not getting set).
+     *
+     * Note: RMAPI lib uses open with O_NONBLOCK followed by fcntl F_SETFL
+     * (to clear the O_NONBLOCK flag) for its async gpu attach path on Linux.
+     * So, this is needed for Linux compat.
+     */
+    if ((cmd == FIONBIO) || (cmd == FIOASYNC)) {
+        if (*((int *)data) == 0)
+            return 0;
+    }
+
+    if (__NV_IOC_TYPE(cmd) != NV_IOCTL_MAGIC)
+        return ENOTTY;
+
     size = __NV_IOC_SIZE(cmd);
     nr = __NV_IOC_NR(cmd);
 
@@ -409,6 +383,21 @@ int nvidia_handle_ioctl(
             status = nvidia_get_card_info(args, size);
             break;
 
+        case NV_ESC_WAIT_OPEN_COMPLETE:
+        {
+            if (nv->flags & NV_FLAG_CONTROL) {
+                status = EINVAL;
+                break;
+            }
+
+            /*
+             * Opens are not deferred to a background thread on this platform
+             * (even with O_NONBLOCK). So, we may assume success.
+             */
+            status = 0;
+            break;
+        }
+
         case NV_ESC_ATTACH_GPUS_TO_FD:
         {
             size_t num_arg_gpus = size / sizeof(NvU32);
@@ -534,6 +523,11 @@ void nvidia_close_ctl(
         nv->flags &= ~NV_FLAG_OPEN;
     }
     nv_unlock_api(nv);
+
+    if (nvffp->mmap_context.valid && nvffp->mmap_context.memArea.pRanges != NULL)
+    {
+        os_free_mem(&nvffp->mmap_context.memArea.pRanges);
+    }
 }
 
 static void nv_dev_free_stacks(nvidia_softc_t *sc)
@@ -599,30 +593,14 @@ int nvidia_open_dev(
     sp = sc->sp[NV_DEV_STACK_API];
 
     if (sc->refcnt == 0) {
+        rm_set_rm_firmware_requested(sp, nv);
+
         if (!rm_init_adapter(sp, nv)) {
             device_printf(sc->dev, "NVRM: rm_init_adapter() failed!\n");
             status = EIO;
             goto failed;
         }
 
-        if (nv->ud.size != 0) {
-            sc->UD_sg_list = sglist_alloc(1, M_WAITOK);
-            if (!sc->UD_sg_list)
-                goto failed;
-
-            sglist_append_phys(sc->UD_sg_list, nv->ud.cpu_address, nv->ud.size);
-
-            sc->UD_object = vm_pager_allocate(OBJT_SG, sc->UD_sg_list,
-                    nv->ud.size, (VM_PROT_READ | VM_PROT_WRITE),
-                    0, NULL);
-            if (!sc->UD_object)
-                goto failed;
-
-            VM_OBJECT_WLOCK(sc->UD_object);
-            vm_object_set_memattr(sc->UD_object, VM_MEMATTR_UNCACHEABLE);
-            VM_OBJECT_WUNLOCK(sc->UD_object);
-        }
-
         nv->flags |= NV_FLAG_OPEN;
     }
 
@@ -632,17 +610,8 @@ int nvidia_open_dev(
     return 0;
 
 failed:
-    if (sc->refcnt == 0) {
-        if (sc->UD_object != NULL)
-            NV_VM_OBJECT_DEALLOCATE(sc->UD_object);
-        if (sc->UD_sg_list != NULL)
-            NV_SGLIST_FREE(sc->UD_sg_list);
-
-        if (status != EIO)
-            NV_SHUTDOWN_ADAPTER(sp, nv);
-
-        nv_dev_free_stacks(sc);
-    }
+    NV_SHUTDOWN_ADAPTER(sp, nv);
+    nv_dev_free_stacks(sc);
     nv_unlock_api(nv);
 
     return status;
@@ -680,11 +649,6 @@ void nvidia_close_dev(
     if (--sc->refcnt == 0) {
         nvidia_stack_t *sp = sc->sp[NV_DEV_STACK_API];
 
-        if (sc->UD_object != NULL)
-            NV_VM_OBJECT_DEALLOCATE(sc->UD_object);
-        if (sc->UD_sg_list != NULL)
-            NV_SGLIST_FREE(sc->UD_sg_list);
-
         NV_SHUTDOWN_ADAPTER(sp, nv);
 
         nv_dev_free_stacks(sc);
@@ -983,7 +947,7 @@ nv_file_private_t* NV_API_CALL nv_get_file_private(
         return NULL;
 
     p = pFile->f_cdevpriv;
-    if(!p)
+    if (!p)
         goto failed;
 
     if (ctl)
@@ -1060,6 +1024,8 @@ NV_STATUS NV_API_CALL nv_add_mapping_context_to_file(
         goto done;
     }
 
+    os_mem_set((void*) nvamc, 0, sizeof(nv_alloc_mapping_context_t));
+
     if (NV_IS_CTL_DEVICE(nv))
     {
         nvamc->alloc = pAllocPriv;
@@ -1067,14 +1033,26 @@ NV_STATUS NV_API_CALL nv_add_mapping_context_to_file(
     }
     else
     {
-        nvamc->mmap_start = nvuap->mmap_start;
-        nvamc->mmap_size = nvuap->mmap_size;
+        status = os_alloc_mem((void**) &nvamc->memArea.pRanges,
+            sizeof(MemoryRange) * nvuap->memArea.numRanges);
+
+        if (status != NV_OK)
+        {
+            nvamc->memArea.pRanges = NULL;
+            goto done;
+        }
+
+        nvamc->memArea.numRanges = nvuap->memArea.numRanges;
+        os_mem_copy(nvamc->memArea.pRanges, nvuap->memArea.pRanges,
+            sizeof(MemoryRange) * nvuap->memArea.numRanges);
+
         nvamc->access_start = nvuap->access_start;
         nvamc->access_size = nvuap->access_size;
         nvamc->remap_prot_extra = nvuap->remap_prot_extra;
     }
 
     nvamc->prot = prot;
+    nvamc->caching = nvuap->caching;
     nvamc->valid = NV_TRUE;
 
 done:
@@ -1133,6 +1111,27 @@ static void flush_cpu_cache(void)
 }
 #endif
 
+static inline vm_memattr_t
+_convert_caching_to_attr
+(
+    NvU32 caching
+)
+{
+    switch (caching) {
+        case NV_MEMORY_UNCACHED:
+            return VM_MEMATTR_UNCACHEABLE;
+        case NV_MEMORY_DEFAULT:
+        case NV_MEMORY_CACHED:
+            return VM_MEMATTR_WRITE_BACK;
+        case NV_MEMORY_UNCACHED_WEAK:
+            return VM_MEMATTR_WEAK_UNCACHEABLE;
+        case NV_MEMORY_WRITECOMBINED:
+            return VM_MEMATTR_WRITE_COMBINING;
+    }
+
+    return VM_MEMATTR_UNCACHEABLE;
+}
+
 static NvS32 nv_alloc_contig_pages(
     NvU32       count,
     NvU32       cache_type,
@@ -1149,25 +1148,7 @@ static NvS32 nv_alloc_contig_pages(
     int flags = (zero ? M_ZERO : 0);
     int status;
 
-    switch (cache_type) {
-        case NV_MEMORY_UNCACHED:
-            attr = VM_MEMATTR_UNCACHEABLE;
-            break;
-        case NV_MEMORY_DEFAULT:
-        case NV_MEMORY_CACHED:
-            attr = VM_MEMATTR_WRITE_BACK;
-            break;
-        case NV_MEMORY_UNCACHED_WEAK:
-            attr = VM_MEMATTR_WEAK_UNCACHEABLE;
-            break;
-        case NV_MEMORY_WRITECOMBINED:
-            attr = VM_MEMATTR_WRITE_COMBINING;
-            break;
-        default:
-            nv_printf(NV_DBG_ERRORS,
-                  "NVRM: unknown mode in nv_alloc_contig_pages()\n");
-            return EINVAL;
-    }
+    attr = _convert_caching_to_attr(cache_type);
 
     at = malloc(sizeof(nvidia_alloc_t), M_NVIDIA, (M_WAITOK | M_ZERO));
     if (!at)
@@ -1281,25 +1262,7 @@ static NvS32 nv_alloc_system_pages(
     int flags = (zero ? M_ZERO : 0);
     int status;
 
-    switch (cache_type) {
-        case NV_MEMORY_UNCACHED:
-            attr = VM_MEMATTR_UNCACHEABLE;
-            break;
-        case NV_MEMORY_DEFAULT:
-        case NV_MEMORY_CACHED:
-            attr = VM_MEMATTR_WRITE_BACK;
-            break;
-        case NV_MEMORY_UNCACHED_WEAK:
-            attr = VM_MEMATTR_WEAK_UNCACHEABLE;
-            break;
-        case NV_MEMORY_WRITECOMBINED:
-            attr = VM_MEMATTR_WRITE_COMBINING;
-            break;
-        default:
-            nv_printf(NV_DBG_ERRORS,
-                  "NVRM: unknown mode in nv_alloc_system_pages()\n");
-            return EINVAL;
-    }
+    attr = _convert_caching_to_attr(cache_type);
 
     at = malloc(sizeof(nvidia_alloc_t), M_NVIDIA, (M_WAITOK | M_ZERO));
     if (!at)
@@ -1410,10 +1373,12 @@ static NvS32 nv_free_system_pages(
 NV_STATUS NV_API_CALL nv_alias_pages(
     nv_state_t *nv,
     NvU32       count,
+    NvU64       page_size,
     NvU32       alloc_type_contiguous,
     NvU32       cache_type,
     NvU64       guest_id,
     NvU64      *pte_array,
+    NvBool      carveout,
     void      **priv_data
 )
 {
@@ -1423,6 +1388,7 @@ NV_STATUS NV_API_CALL nv_alias_pages(
 NV_STATUS NV_API_CALL nv_alloc_pages(
     nv_state_t *nv,
     NvU32       count,
+    NvU64       page_size,
     NvBool      alloc_type_contiguous,
     NvU32       cache_type,
     NvBool      alloc_type_zeroed,
@@ -1535,7 +1501,7 @@ int nvidia_mmap_ctl_single(
      * If no mmap context exists on this file descriptor, this mapping wasn't
      * previously validated with the RM so it must be rejected.
      */
-    if (!mmap_context->valid) {
+    if (!mmap_context->valid || *offset != 0) {
         nv_printf(NV_DBG_ERRORS, "NVRM: VM: invalid mmap\n");
         status = EINVAL;
         goto done;
@@ -1569,13 +1535,11 @@ int nvidia_mmap_dev_single(
     int nprot
 )
 {
-    nv_state_t *nv;
-    struct nvidia_softc *sc;
     nv_alloc_mapping_context_t *mmap_context = &nvffp->mmap_context;
     int status;
-
-    nv = nvffp->nv;
-    sc = nv->os_state;
+    struct sglist *pSgList;
+    NvU64 idx = 0;
+    NvU64 tot = 0;
 
     sx_xlock(&nvffp->fops_sx);
 
@@ -1589,33 +1553,41 @@ int nvidia_mmap_dev_single(
         goto done;
     }
 
-    *offset = mmap_context->mmap_start;
-    size = mmap_context->mmap_size;
-
     if ((mmap_context->prot & nprot) != nprot) {
         status = EACCES;
         goto done;
     }
 
-    if (IS_UD_OFFSET(nv, *offset, size)) {
-        *object = sc->UD_object;
-        vm_object_reference(*object);
-        *offset = (*offset - nv->ud.cpu_address);
-        status = 0;
-    } else if (IS_FB_OFFSET(nv, *offset, size)) {
-        *object = sc->BAR_objects[NV_GPU_BAR_INDEX_FB];
-        vm_object_reference(*object);
-        *offset = (*offset - nv->fb->cpu_address);
-        status = 0;
-    } else if (IS_REG_OFFSET(nv, *offset, size)) {
-        *object = sc->BAR_objects[NV_GPU_BAR_INDEX_REGS];
-        vm_object_reference(*object);
-        *offset = (*offset - nv->regs->cpu_address);
-        status = 0;
-    } else {
-        status = EINVAL;
+    status = EINVAL;
+
+    pSgList = sglist_alloc(mmap_context->memArea.numRanges, M_WAITOK);
+    if (pSgList == NULL)
+    {
+        goto done;
     }
 
+    for (idx = 0; idx < mmap_context->memArea.numRanges; idx++)
+    {
+        sglist_append_phys(pSgList, mmap_context->memArea.pRanges[idx].start, mmap_context->memArea.pRanges[idx].size);
+        tot += mmap_context->memArea.pRanges[idx].size;
+    }
+
+    *object = vm_pager_allocate(OBJT_SG, pSgList, tot, nprot, 0, NULL);
+
+    if (*object == NULL)
+    {
+        goto free_sglist;
+    }
+
+    VM_OBJECT_WLOCK(*object);
+    vm_object_set_memattr(*object, _convert_caching_to_attr(mmap_context->caching));
+    VM_OBJECT_WUNLOCK(*object);
+
+    status = NV_OK;
+
+free_sglist:
+    // Free sglist regardless here, now it is owned by the pager if allocation has succeeded.
+    NV_SGLIST_FREE(pSgList);
 done:
     sx_xunlock(&nvffp->fops_sx);
 
@@ -1707,28 +1679,6 @@ nv_state_t* NV_API_CALL nv_get_ctl_state(void)
     return &nvidia_ctl_state;
 }
 
-NV_STATUS NV_API_CALL nv_dma_map_pages(
-    nv_dma_device_t *dma_dev,
-    NvU64            page_count,
-    NvU64           *pte_array,
-    NvBool           contig,
-    NvU32            cache_type,
-    void           **priv
-)
-{
-    return NV_ERR_NOT_SUPPORTED;
-}
-
-NV_STATUS NV_API_CALL nv_dma_unmap_pages(
-    nv_dma_device_t *dma_dev,
-    NvU64            page_count,
-    NvU64           *pte_array,
-    void           **priv
-)
-{
-    return NV_ERR_NOT_SUPPORTED;
-}
-
 NV_STATUS NV_API_CALL nv_dma_map_alloc(
     nv_dma_device_t *dma_dev,
     NvU64            page_count,
@@ -1793,12 +1743,6 @@ void NV_API_CALL nv_dma_cache_invalidate(
 {
 }
 
-void NV_API_CALL nv_dma_enable_nvlink(
-    nv_dma_device_t *dma_dev
-)
-{
-}
-
 NV_STATUS NV_API_CALL nv_log_error(
     nv_state_t *nv,
     NvU32       error_number,
@@ -1809,13 +1753,6 @@ NV_STATUS NV_API_CALL nv_log_error(
     return NV_OK;
 }
 
-NvU64 NV_API_CALL nv_get_dma_start_address(
-    nv_state_t *nv
-)
-{
-    return 0;
-}
-
 NV_STATUS NV_API_CALL nv_set_primary_vga_status(
     nv_state_t *nv
 )
@@ -1823,13 +1760,6 @@ NV_STATUS NV_API_CALL nv_set_primary_vga_status(
     return NV_ERR_NOT_SUPPORTED;
 }
 
-NV_STATUS NV_API_CALL nv_pci_trigger_recovery(
-    nv_state_t *nv
-)
-{
-    return NV_ERR_NOT_SUPPORTED;
-}
-
 NvBool NV_API_CALL nv_requires_dma_remap(
     nv_state_t *nv
 )
@@ -1919,6 +1849,7 @@ NV_STATUS NV_API_CALL nv_get_device_memory_config(
     nv_state_t *nv,
     NvU64 *pAddrSysPhys,
     NvU64 *pAddrGuestPhys,
+    NvU64 *rsvd_phys,
     NvU32 *pAddrWidth,
     NvS32 *pNodeId
 )
@@ -2053,30 +1984,6 @@ NV_STATUS NV_API_CALL nv_get_phys_pages(
     return NV_OK;
 }
 
-NV_STATUS NV_API_CALL nv_get_ibmnpu_genreg_info(
-    nv_state_t *nv,
-    NvU64      *addr,
-    NvU64      *size,
-    void      **device
-)
-{
-    return NV_ERR_NOT_SUPPORTED;
-}
-
-NV_STATUS NV_API_CALL nv_get_ibmnpu_relaxed_ordering_mode(
-    nv_state_t *nv,
-    NvBool *mode
-)
-{
-    return NV_ERR_NOT_SUPPORTED;
-}
-
-void NV_API_CALL nv_wait_for_ibmnpu_rsync(
-    nv_state_t *nv
-)
-{
-}
-
 void NV_API_CALL nv_p2p_free_platform_data(
     void *data
 )
@@ -2200,7 +2107,6 @@ NV_STATUS NV_API_CALL nv_dma_import_dma_buf
     nv_dma_device_t *dma_dev,
     struct dma_buf *dma_buf,
     NvU32 *size,
-    void **user_pages,
     struct sg_table **sgt,
     nv_dma_buf_t **import_priv
 )
@@ -2213,7 +2119,6 @@ NV_STATUS NV_API_CALL nv_dma_import_from_fd
     nv_dma_device_t *dma_dev,
     NvS32 fd,
     NvU32 *size,
-    void **user_pages,
     struct sg_table **sgt,
     nv_dma_buf_t **import_priv
 )
@@ -2231,7 +2136,6 @@ void NV_API_CALL nv_dma_release_sgt
 
 void NV_API_CALL nv_dma_release_dma_buf
 (
-    void *user_pages,
     nv_dma_buf_t *import_priv
 )
 {
@@ -2290,6 +2194,17 @@ nv_schedule_uvm_isr(nv_state_t *nv)
 {
 }
 
+NV_STATUS NV_API_CALL
+nv_schedule_uvm_drain_p2p(NvU8 *pUuid)
+{
+    return NV_ERR_NOT_SUPPORTED;
+}
+
+void NV_API_CALL
+nv_schedule_uvm_resume_p2p(NvU8 *pUuid)
+{
+}
+
 NvBool NV_API_CALL nv_is_rm_firmware_active(
     nv_state_t *nv
 )
@@ -2305,13 +2220,28 @@ const void* NV_API_CALL nv_get_firmware(
     NvU32 *fw_size
 )
 {
-    return NULL;
+    const struct firmware *fw;
+
+    const char *fw_name = nv_firmware_for_chip_family(fw_type, fw_chip_family);
+    if (fw_name == NULL)
+        return NULL;
+
+    fw = firmware_get(fw_name);
+    if (fw == NULL)
+        return NULL;
+
+    *fw_buf = fw->data;
+    *fw_size  = fw->datasize;
+
+    return fw;
 }
 
 void NV_API_CALL nv_put_firmware(
     const void *fw_handle
 )
 {
+    if (fw_handle != NULL)
+        firmware_put(fw_handle, FIRMWARE_UNLOAD);
 }
 
 NvBool NV_API_CALL nv_is_chassis_notebook(void)
@@ -2331,11 +2261,6 @@ void NV_API_CALL nv_disallow_runtime_suspend(
 {
 }
 
-NvU32 NV_API_CALL nv_get_os_type(void)
-{
-    return OS_TYPE_FREEBSD;
-}
-
 void NV_API_CALL nv_flush_coherent_cpu_cache_range
 (
     nv_state_t *nv,
@@ -2381,3 +2306,98 @@ void NV_API_CALL nv_destroy_nano_timer(
 )
 {
 }
+
+void NV_API_CALL nv_get_disp_smmu_stream_ids(
+    nv_state_t *nv,
+    NvU32      *dispIsoStreamId,
+    NvU32      *dispNisoStreamId
+)
+{
+    /* NV_U32_MAX is used to indicate that the platform does not support SMMU */
+    *dispIsoStreamId = NV_U32_MAX;
+    *dispNisoStreamId = NV_U32_MAX;
+}
+
+void NV_API_CALL nv_get_screen_info(
+    nv_state_t  *nv,
+    NvU64       *pPhysicalAddress,
+    NvU32       *pFbWidth,
+    NvU32       *pFbHeight,
+    NvU32       *pFbDepth,
+    NvU32       *pFbPitch,
+    NvU64       *pFbSize
+)
+{
+    /*
+     * Look up EFI framebuffer information passed to the FreeBSD kernel by the
+     * bootloader.
+     *
+     * Adapted from a suggestion by Conrad Meyer <cem@freebsd.org>.
+     */
+    caddr_t kmdp = preload_search_by_type("elf kernel") ?:
+                   preload_search_by_type("elf64 kernel");
+
+    if (kmdp != NULL)
+    {
+        const struct efi_fb *efifb =
+            (const struct efi_fb *)preload_search_info(kmdp, MODINFO_METADATA |
+                                                             MODINFOMD_EFI_FB);
+
+        /* Make sure base address is mapped to GPU BAR */
+        if ((efifb != NULL) && NV_IS_CONSOLE_MAPPED(nv, efifb->fb_addr))
+        {
+            unsigned int depth = fls(efifb->fb_mask_red | efifb->fb_mask_green |
+                            efifb->fb_mask_blue | efifb->fb_mask_reserved);
+            unsigned int bpp = roundup2(depth, NBBY);
+
+            *pPhysicalAddress = efifb->fb_addr;
+            *pFbWidth = efifb->fb_width;
+            *pFbHeight = efifb->fb_height;
+            *pFbDepth = depth;
+            /* fb_stride is in pixels. Convert to bytes */
+            *pFbPitch = efifb->fb_stride * (bpp / NBBY);
+            *pFbSize = (NvU64)(*pFbHeight) * (NvU64)(*pFbPitch);
+            return;
+        }
+    }
+    {
+        const sc_softc_t *sc = sc_get_softc(0, SC_KERNEL_CONSOLE);
+
+        if (sc)
+        {
+            const video_adapter_t *adp = sc->adp;
+
+            if (adp)
+            {
+                const struct video_info *vi = &adp->va_info;
+
+                /* Make sure base address is mapped to GPU BAR */
+                if (vi && (vi->vi_flags & V_INFO_LINEAR) &&
+                    NV_IS_CONSOLE_MAPPED(nv, vi->vi_buffer))
+                {
+                    *pPhysicalAddress = vi->vi_buffer;
+                    *pFbWidth = vi->vi_width;
+                    *pFbHeight = vi->vi_height;
+                    *pFbDepth = vi->vi_depth;
+                    *pFbPitch = adp->va_line_width;
+                    *pFbSize = (NvU64)(*pFbHeight) * (NvU64)(*pFbPitch);
+                    return;
+                }
+            }
+        }
+    }
+
+    *pPhysicalAddress = 0;
+    *pFbWidth = *pFbHeight = *pFbDepth = *pFbPitch = *pFbSize = 0;
+}
+
+NV_STATUS nv_get_syncpoint_aperture
+(
+    NvU32 syncpointId,
+    NvU64 *physAddr,
+    NvU64 *limit,
+    NvU32 *offset
+)
+{
+    return NV_ERR_NOT_SUPPORTED;
+}
